#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DarkShallFall
# 	Widgets, Bitmaps, Animations, Y and X offsets, layers, blending, and model.
# revision: 3		author: -DeToX-
# 	X and Y Scale
# revision: 4		author: Lord Zedd
# 	Fixed up, added flags and types
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


scn3_data_source = Struct("data_source",
    h3_dependency("data_source"),
    ENDIAN=">", SIZE=16
    )


scn3_group_widget_list_widget_list_widget_item = Struct("list_widget_item",
    Bool32("flags", *unknown_flags_32),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_offset"),
    SInt16("widescreen_x_offset"),
    SInt16("widescreen_y_unknown"),
    SInt16("widescreen_x_unknown"),
    SInt16("standard_y_offset"),
    SInt16("standard_x_offset"),
    SInt16("standard_y_unknown"),
    SInt16("standard_x_unknown"),
    h3_dependency("animation"),
    h3_string_id("target"),
    ENDIAN=">", SIZE=48
    )


scn3_group_widget_list_widget = Struct("list_widget",
    h3_dependency("ancestor"),
    Bool32("flags",
        ("horizontal", 1 << 4),
        "loops",
        ),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_offset"),
    SInt16("widescreen_x_offset"),
    SInt16("widescreen_y_unknown"),
    SInt16("widescreen_x_unknown"),
    SInt16("standard_y_offset"),
    SInt16("standard_x_offset"),
    SInt16("standard_y_unknown"),
    SInt16("standard_x_unknown"),
    h3_dependency("animation"),
    h3_string_id("data_source_name"),
    h3_dependency("skin"),
    SInt32("row_count"),
    h3_reflexive("list_widget_items", scn3_group_widget_list_widget_list_widget_item),
    h3_dependency("up_arrow_bitmap"),
    h3_dependency("down_arrow_bitmap"),
    ENDIAN=">", SIZE=128
    )


scn3_group_widget_text_widget = Struct("text_widget",
    h3_dependency("ancestor"),
    Bool32("flags",
        ("left_alignment", 1 << 3),
        "right_alignment",
        ("all_caps", 1 << 6),
        ("wrap_at_bounds", 1 << 13),
        ),
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_string_id("data_source_name"),
    h3_string_id("text_string"),
    h3_string_id("text_color"),
    SInt16("text_font"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=76
    )


scn3_group_widget_bitmap_widget = Struct("bitmap_widget",
    h3_dependency("ancestor"),
    Bool32("flags",
        ("scale_to_bounds", 1 << 3),
        "replace_with_blur",
        ("replace_with_white", 1 << 10),
        "replace_with_black",
        ),
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_dependency("bitmap"),
    h3_dependency("unknown_1"),
    SEnum16("blend_method", *bmp3_blend_method),
    SInt16("unknown_2", VISIBLE=False),
    SInt16("sprite_index"),
    SInt16("unknown_3", VISIBLE=False),
    h3_string_id("data_source_name"),
    h3_string_id("sprite_data_source_name"),
    ENDIAN=">", SIZE=108
    )


scn3_group_widget_model_widget_camera_refinement_zoom_data_1 = Struct("zoom_data_1",
    h3_rawdata_ref("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


scn3_group_widget_model_widget_camera_refinement_zoom_data_2 = Struct("zoom_data_2",
    h3_rawdata_ref("unknown", VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


scn3_group_widget_model_widget_camera_refinement = Struct("camera_refinement",
    h3_string_id("biped", VISIBLE=False),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=8, VISIBLE=False),
    h3_reflexive("zoom_data_1", scn3_group_widget_model_widget_camera_refinement_zoom_data_1),
    h3_reflexive("zoom_data_2", scn3_group_widget_model_widget_camera_refinement_zoom_data_2),
    VISIBLE=False,
    ENDIAN=">", SIZE=60
    )


scn3_group_widget_model_widget = Struct("model_widget",
    h3_dependency("ancestor"),
    Bool32("flags", *unknown_flags_32),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_reflexive("camera_refinement", scn3_group_widget_model_widget_camera_refinement),
    ENDIAN=">", SIZE=72
    )


scn3_group_widget = Struct("group_widget",
    h3_dependency("ancestor"),
    Bool32("flags",
        ("initially_hidden", 1 << 3),
        ),
    h3_string_id("name"),
    SInt16("unknown", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_offset"),
    SInt16("widescreen_x_offset"),
    SInt16("widescreen_y_unknown"),
    SInt16("widescreen_x_unknown"),
    SInt16("standard_y_offset"),
    SInt16("standard_x_offset"),
    SInt16("standard_y_unknown"),
    SInt16("standard_x_unknown"),
    h3_dependency("animation"),
    h3_reflexive("list_widgets", scn3_group_widget_list_widget),
    h3_reflexive("text_widgets", scn3_group_widget_text_widget),
    h3_reflexive("bitmap_widgets", scn3_group_widget_bitmap_widget),
    h3_reflexive("model_widgets", scn3_group_widget_model_widget),
    ENDIAN=">", SIZE=108
    )


scn3_button_key_legend = Struct("button_key_legend",
    h3_dependency("legend"),
    ENDIAN=">", SIZE=16
    )


scn3_body = Struct("tagdata",
    Bool32("flags", *unknown_flags_32),
    h3_string_id("name"),
    SInt16("unknown_0", VISIBLE=False),
    SInt16("layer"),
    SInt16("widescreen_y_bounds_min"),
    SInt16("widescreen_x_bounds_min"),
    SInt16("widescreen_y_bounds_max"),
    SInt16("widescreen_x_bounds_max"),
    SInt16("standard_y_bounds_min"),
    SInt16("standard_x_bounds_min"),
    SInt16("standard_y_bounds_max"),
    SInt16("standard_x_bounds_max"),
    h3_dependency("animation"),
    h3_dependency("strings"),
    h3_dependency("ancestor"),
    h3_string_id("default_key_legend_string"),
    h3_reflexive("data_sources", scn3_data_source),
    h3_reflexive("group_widgets", scn3_group_widget),
    h3_reflexive("button_key_legends", scn3_button_key_legend),
    h3_dependency("ui_sounds"),
    ascii_str32("script_title"),
    SInt16("script_index"),
    SInt16("unknown_1", VISIBLE=False),
    ENDIAN=">", SIZE=168
    )


def get():
    return scn3_def

scn3_def = TagDef("scn3",
    h3_blam_header('scn3'),
    scn3_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["scn3"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.weap import *
from .obje import *
from .item import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=2)

weap_body = dict(weap_body)
weap_body[0] = obje_attrs

def get():
    return weap_def

weap_def = TagDef("weap",
    blam_header('weap', 2),
    weap_body,

    ext=".weapon", endian=">", tag_cls=WeapTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DarkShallFall
# 	Labeled about 40 values, structures, and idents.
# revision: 3		author: -DeToX- and DarkShallFall
# 	Labelled reflexives and found their purposes, aswell as the service tag bipd strings.
# revision: 4		author: Lord Zedd
# 	Got most things
# revision: 5		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

wigl_alert_icon = (
    "none",
    "download",
    "pause",
    "upload",
    "checkbox",
    )

wigl_dialog_default_option = (
    "option_1",
    "option_2",
    "option_3",
    "option_4",
    )


wigl_text_color = Struct("text_color",
    h3_string_id("name"),
    color_argb_float("color"),
    ENDIAN=">", SIZE=20
    )


wigl_player_color_player_text_color = Struct("player_text_color",
    color_argb_float("color"),
    ENDIAN=">", SIZE=16
    )


wigl_player_color_team_text_color = Struct("team_text_color",
    color_argb_float("color"),
    ENDIAN=">", SIZE=16
    )


wigl_player_color_player_ui_color = Struct("player_ui_color",
    color_argb_float("color"),
    ENDIAN=">", SIZE=16
    )


wigl_player_color_team_ui_color = Struct("team_ui_color",
    color_argb_float("color"),
    ENDIAN=">", SIZE=16
    )


wigl_player_color = Struct("player_color",
    h3_reflexive("player_text_color", wigl_player_color_player_text_color),
    h3_reflexive("team_text_color", wigl_player_color_team_text_color),
    h3_reflexive("player_ui_color", wigl_player_color_player_ui_color),
    h3_reflexive("team_ui_color", wigl_player_color_team_ui_color),
    ENDIAN=">", SIZE=48
    )


wigl_alert = Struct("alert",
    h3_string_id("name"),
    Bool8("flags", *unknown_flags_8),
    SInt8("unknown_0"),
    SEnum8("icon", *wigl_alert_icon),
    SInt8("unknown_1", VISIBLE=False),
    h3_string_id("title"),
    h3_string_id("body"),
    ENDIAN=">", SIZE=16
    )


wigl_dialog = Struct("dialog",
    h3_string_id("name"),
    SInt16("unknown_0"),
    SInt16("unknown_1", VISIBLE=False),
    h3_string_id("title"),
    h3_string_id("body"),
    Array("option_array", SUB_STRUCT=h3_string_id("option"), SIZE=4),
    h3_string_id("key_legend"),
    SEnum16("default_option", *wigl_dialog_default_option),
    SInt16("unknown_2", VISIBLE=False),
    ENDIAN=">", SIZE=40
    )


wigl_global_data_source = Struct("global_data_source",
    h3_dependency("data_source"),
    ENDIAN=">", SIZE=16
    )


wigl_body = Struct("tagdata",
    SInt16("inc_text_update_period"),
    SInt16("inc_text_block_character"),
    Float("near_clip_plane_distance"),
    Float("projection_plane_distance"),
    Float("far_clip_plane_distance"),
    h3_dependency("global_strings"),
    h3_dependency("damage_type_strings"),
    h3_dependency("main_menu_music"),
    SInt32("music_fade_time"),
    color_argb_float("color"),
    color_argb_float("text_stroke_color"),
    h3_reflexive("text_colors", wigl_text_color),
    h3_reflexive("player_colors", wigl_player_color),
    h3_dependency("ui_sounds"),
    h3_reflexive("alerts", wigl_alert),
    h3_reflexive("dialogs", wigl_dialog),
    h3_reflexive("global_data_sources", wigl_global_data_source),
    QStruct("widescreen_bitmap_scale", INCLUDE=xy_float),
    QStruct("standard_bitmap_scale", INCLUDE=xy_float),
    QStruct("menu_blur", INCLUDE=xy_float),
    ascii_str32("ui_spartan_biped_name"),
    ascii_str32("ui_spartan_ai_squad_name"),
    h3_string_id("ui_spartan_ai_location_name"),
    ascii_str32("ui_elite_biped_name"),
    ascii_str32("ui_elite_ai_squad_name"),
    h3_string_id("ui_elite_ai_location_name"),
    SInt32("single_scroll_speed"),
    SInt32("scroll_speed_transition_wait_time"),
    SInt32("held_scroll_speed"),
    SInt32("attract_video_idle_wait"),
    ENDIAN=">", SIZE=352
    )


def get():
    return wigl_def

wigl_def = TagDef("wigl",
    h3_blam_header('wigl'),
    wigl_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wigl"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import zlib


def decompress_lzx_stream(stream, start, comp_size, uncomp_offset, uncomp_size):
    stream.seek(start)
    comp_data = stream.read(comp_size)
    return comp_data


def decompress_deflate_stream(stream, start, comp_size, uncomp_offset, uncomp_size):
    stream.seek(start)
    comp_data = stream.read(comp_size)
    decompressor = zlib.decompressobj(-15)
    uncomp_data = decompressor.decompress(comp_data, uncomp_offset + uncomp_size)
    return uncomp_data[uncomp_offset: ]


class RawdataManager:
    # tags with zone references:
    # bitm, jmad, mode, pmdf, sLdT, sbsp, snd!
    _parent_halo_map = None

    _zone_meta = None
    _play_meta = None

    _map_index_to_name = ()

    def __init__(self, parent_halo_map):
        self._parent_halo_map = parent_halo_map
        self._map_index_to_name = {}
        if parent_halo_map:
            self._zone_meta = parent_halo_map.get_root_tag(
                "cache_file_resource_gestalt")
            self._play_meta = parent_halo_map.get_root_tag(
                "cache_file_resource_layout_table")

    def add_shared_map_name(self, ext_cache_name, map_name):
        assert isinstance(map_name, str)
        if not self._play_meta:
            return

        ext_cache_refs = self._play_meta.external_cache_references.STEPTREE
        for i in range(len(ext_cache_refs)):
            curr_ext_cache_name = ext_cache_refs[i].map_path.lower().\
                                  replace("\\", "/").split("/")[-1].\
                                  split(".")[0]
            if curr_ext_cache_name == ext_cache_name:
                self._map_index_to_name[i] = map_name
                break

    def get_halo_map(self, map_index):
        if map_index < 0:
            return self._parent_halo_map
        elif self._parent_halo_map:
            return self._parent_halo_map.maps.get(
                self._map_index_to_name.get(map_index))

    def get_tag_resource_data(self, tag_rsrc_index, *chunk_sizes):
        tag_rsrc_datas = []
        tag_rsrc_info = self.get_tag_resource(tag_rsrc_index)
        if not tag_rsrc_info:
            return tag_rsrc_datas

        segment_info = self.get_segment_info(tag_rsrc_info.segment_index)
        chunk_sizes += (None, ) * (len(segment_info) - len(chunk_sizes))
        chunk_sizes = list(chunk_sizes)
        for raw_page_index, chunk_offset, seg_sizes in segment_info:
            raw_page_data = self.get_raw_page_data(raw_page_index, chunk_offset,
                                                   chunk_sizes.pop(0))
            if raw_page_data is None: continue

            # ignoring seg_sizes for now
            tag_rsrc_data = raw_page_data

            tag_rsrc_datas.append(tag_rsrc_data)

        return tag_rsrc_datas

    def get_raw_page_data(self, raw_page_index, chunk_offset=0, chunk_size=None):
        raw_page = self.get_raw_page(raw_page_index)
        if raw_page is None: return None

        halo_map = self.get_halo_map(raw_page.shared_cache_index)
        if halo_map is None: return None

        map_data = halo_map.map_data
        data_offset = halo_map.map_pointer_converter.v_ptr_to_f_ptr(
            raw_page.block_offset)
        comp_size = raw_page.compressed_block_size
        if chunk_size is None:
            chunk_size = raw_page.uncompressed_block_size

        if raw_page.compression_codec.enum_name == "NONE":
            map_data.seek(data_offset + chunk_offset)
            uncomp_data = map_data.read(chunk_size)
        elif raw_page.compression_codec.enum_name == "deflate":
            uncomp_data = decompress_deflate_stream(
                map_data, data_offset, comp_size, chunk_offset, chunk_size)
        elif raw_page.compression_codec.enum_name == "lzx":
            uncomp_data = decompress_lzx_stream(
                map_data, data_offset, comp_size, chunk_offset, chunk_size)
        else:
            raise ValueError("Unknown compression codec '%s'" %
                             raw_page.compression_codec.data)

        if len(uncomp_data) < chunk_size:
            # pad the data up to the requested size
            uncomp_data += b'\x00' * (chunk_size - len(uncomp_data))

        return uncomp_data

    def get_tag_resource_fixup(self, tag_rsrc_index):
        tag_rsrc_info = self.get_tag_resource(tag_rsrc_index)
        if not tag_rsrc_info:
            return (b'', (), ())

        fixup_chunk = self._zone_meta.fixup_info.data[
            tag_rsrc_info.fixup_info_offset:
            tag_rsrc_info.fixup_info_offset + tag_rsrc_info.fixup_info_size
            ]

        return (fixup_chunk, tag_rsrc_info.resource_fixups.STEPTREE,
                tag_rsrc_info.resource_definition_fixups.STEPTREE)

    def get_segment_info(self, segment_index):
        if self._play_meta and isinstance(segment_index, int):
            raw_sizes = self._play_meta.raw_sizes.STEPTREE
            segments = self._play_meta.segments.STEPTREE
            if segment_index in range(len(segments)):
                segment = segments[segment_index]
                page_infos = []
                for i in range(len(segment.page_indices)):
                    page_info = [-1, 0, []]
                    page_infos.append(page_info)

                    raw_page_index = segment.page_indices[i]
                    raw_page = self.get_raw_page(raw_page_index)
                    if not raw_page:
                        continue

                    seg_sizes = []
                    page_info[:] = (
                        raw_page_index,
                        segment.segment_offsets[i],
                        seg_sizes
                        )
                    size_index = segment.raw_size_indices[i]
                    if size_index == -1:
                        seg_sizes.append((0, raw_page.uncompressed_block_size))
                    else:
                        for chunk in seg_sizes[size_index].chunks.STEPTREE:
                            seg_sizes.append((chunk[0], chunk[1]))

                return page_infos

        return ()

    def get_raw_page(self, page_index):
        if self._play_meta and isinstance(page_index, int):
            raw_pages = self._play_meta.raw_pages.STEPTREE
            if page_index in range(len(raw_pages)):
                return raw_pages[page_index]

    def get_tag_resource(self, tag_rsrc_index):
        if self._zone_meta and isinstance(tag_rsrc_index, int):
            tag_resources = self._zone_meta.tag_resources.STEPTREE
            if tag_rsrc_index in range(len(tag_resources)):
                return tag_resources[tag_rsrc_index]



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


udlg_vocalization = Struct("vocalization",
    Bool16("flags",
        "new_vocalization",
        ),
    SInt16("unknown", VISIBLE=False),
    h3_string_id("vocalization"),
    h3_dependency("sound"),
    ENDIAN=">", SIZE=24
    )


udlg_body = Struct("tagdata",
    h3_dependency("global_dialogue_info"),
    Bool32("flags",
        "female",
        ),
    h3_reflexive("vocalizations", udlg_vocalization),
    h3_string_id("mission_dialogue_designator"),
    ENDIAN=">", SIZE=36
    )


def get():
    return udlg_def

udlg_def = TagDef("udlg",
    h3_blam_header('udlg'),
    udlg_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["udlg"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

'''
Used for extracting tag definitions to a format that can be easily read.
'''
from reclaimer.field_types import Reflexive
from reclaimer.h2.field_types import H2Reflexive
from traceback import format_exc
from supyr_struct.tag import Tag
from supyr_struct.field_types import Union


def find_enums(block, next_blocks):
    typ = block.TYPE
    if "enum" in typ.name.lower():
        next_blocks.append(block)
        return

    if typ.is_array:
        if block.get_desc('TYPE', 'SUB_STRUCT').is_block:
            if len(block) == 0:
                block.append()
            find_enums(block[0], next_blocks)
    elif typ == Union:
        for case in block.CASE_MAP:
            block.set_active(case)
            find_enums(block.u_node, next_blocks)
        block.set_active()
    elif typ.is_container or typ.is_struct:
        for i in range(len(block)):
            if block.get_desc('TYPE', i).is_block:
                if hasattr(block[i], "TYPE"):
                    find_enums(block[i], next_blocks)

    if (hasattr(block, "STEPTREE") and
            block.get_desc('TYPE', 'STEPTREE').is_block):
        find_enums(block.STEPTREE, next_blocks)


def find_next_blocks(block, next_blocks):
    typ = block.TYPE
    if typ in (Reflexive, H2Reflexive):
        block.STEPTREE.append()
        next_blocks.append(block.STEPTREE[-1])
        return

    if typ.is_array:
        if block.get_desc('TYPE', 'SUB_STRUCT').is_block:
            if len(block) == 0:
                block.append()
            find_next_blocks(block[0], next_blocks)
    elif typ == Union:
        for case in block.CASE_MAP:
            block.set_active(case)
            find_next_blocks(block.u_node, next_blocks)
        block.set_active()
    elif typ.is_container or typ.is_struct:
        for i in range(len(block)):
            if block.get_desc('TYPE', i).is_block:
                if hasattr(block[i], "TYPE"):
                    find_next_blocks(block[i], next_blocks)

    if (hasattr(block, "STEPTREE") and
            block.get_desc('TYPE', 'STEPTREE').is_block):
        find_next_blocks(block.STEPTREE, next_blocks)


def print_tag_def(tag_def):
    # make an empty tag to do w/e with
    new_tag = tag_def.build()

    # write the tag definition to a file
    with open("%s_def.txt" % tag_def.def_id, "w") as f:
        tagdata = new_tag
        def_id = tag_def.def_id
        if isinstance(new_tag, Tag):
            tagdata = new_tag.data.tagdata
            blocks = [new_tag.data.blam_header, tagdata]
            f.write("%s (%s)\n\n" % (def_id, tag_def.ext[1:]))
        else:
            blocks = [tagdata]
            f.write("%s\n\n" % def_id)

        f.write("#"*60 + "\n")
        f.write("#" + (" "*20) + "STRUCTURES\n")
        f.write("\n\n")

        # loop over each block until they are exhausted
        while blocks:
            new_blocks = []

            for block in blocks:
                if not hasattr(block, "NAME"):
                    continue

                f.write("\n")
                f.write("\n")
                f.write("#"*60 + "\n")
                f.write("#" + (" "*20) + block.NAME + "\n")
                f.write("\n")
                f.write(block.pprint(show=("name", "type", "offset",
                                           "flags", "size")).\
                        replace("[", "{").replace("]", "}"))

                find_next_blocks(block, new_blocks)

            blocks = new_blocks

        f.write("\n\n")
        f.write("#"*60 + "\n")
        f.write("#" + (" "*20) + "ENUMERATORS\n")
        f.write("\n\n")

        # loop over each block until they are exhausted
        enums = []
        find_enums(tagdata, enums)
        i = j = 0
        enum_strs = []
        enum_strs_map = {}
        enum_names = []
        for enum in enums:
            name = ""
            b = enum
            while hasattr(b, 'NAME') and hasattr(b, 'parent'):
                if not (b.TYPE.is_array or b.NAME in ("tagdata", def_id) or
                        b.TYPE in (Reflexive, H2Reflexive)):
                    name = "%s.%s" % (b.NAME, name)
                b = b.parent
            name = name.rstrip(".")

            if not name:
                name = "UNNAMED_ENUM_%s" % i
                i += 1

            desc = enum.desc
            enum_str = "%s" + ("%s {\n" % enum.TYPE.name)
            for k in range(enum.ENTRIES):
                enum_str += "    %s = %s,\n" % (desc[k]['NAME'],
                                                desc[k]['VALUE'])
            enum_str += "    }\n\n"

            if enum_str in enum_strs_map:
                enum_names[enum_strs_map[enum_str]].append(name)
                continue

            enum_strs_map[enum_str] = j
            enum_strs.append(enum_str)
            enum_names.append([name])
            j += 1

        for i in range(len(enum_names)):
            name = ""
            names = enum_names[i]
            for j in range(len(names)):
                name = "%s = %s" % (names[j], name)
            f.write(enum_strs[i] % name)

tag_defs = {}

#import the tag definition so it can be used for making a blank tag
while not tag_defs:
    engine = input("Type in the definition set to use.\n"
                   "Valid values are hek, os_hek, os_v3_hek, and os_v4_hek.\n"
                   ">>> ")
    engine = engine.strip(" ").lower()
    print()
    classes = input("Type in the four character codes of the tag classes to write.\n"
                    "Use commas to separate each one. Spaces are ignored.\n"
                    "Examples include jpt!, vehi, DeLa, snd!, and bitm.\n"
                    "Type **** if you want to write every available tag class.\n"
                    ">>> ")
    if "****" in classes:
        all_classes = () 
        exec("from reclaimer.%s.defs import __all__ as all_classes" % engine)
        classes = ",".join(all_classes)

    classes = classes.replace("!", "_").replace("#", "_").replace("+", "_").\
              replace(" ", '').split(",")
    fixed_classes = []
    for cls in classes:
        fixed_classes.append(cls + " " * (4 - len(cls)))

    classes = []
    for cls in fixed_classes:
        try:
            def_name = cls
            tag_def = None
            if "_meta" in cls:
                def_name = cls.replace("_meta", "")
            exec("from reclaimer.%s.defs.%s import %s_def as tag_def" %
                 (engine, def_name, cls))
            tag_defs[cls] = tag_def
            classes.append(cls)
        except Exception:
            print("Could not load the %s definition" % cls)
            print(format_exc())

for cls in classes:
    try:
        print_tag_def(tag_defs[cls])
    except Exception:
        print("Could not print the %s definition" % cls)
        print(format_exc())

input("Finished. Hit enter to exit. . .")



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Fixed up the layout, mapped out some values..
# revision: 3		author: DarkShallFall
# 	More unknowns and types.
# revision: 4		author: DarkShallFall
# 	Plugin getting close to complete. A little more effort and we will be there.
# revision: 5		author: Lord Zedd
# 	Overhauled.
# revision: 6		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

chdt_hud_widget_animation_data_animation_1_function = (
    "default",
    "use_input",
    "use_range_input",
    "zero",
    )

chdt_hud_widget_placement_data_anchor = (
    "top_left",
    "top_right",
    "bottom_right",
    "bottom_left",
    "center",
    "top_edge",
    "grenade_a",
    "grenade_b",
    "grenade_c",
    "grenade_d",
    "scoreboard_friendly",
    "scoreboard_enemy",
    "health_and_shield",
    "bottom_edge",
    "unknown_0",
    "equipment",
    "unknown_1",
    "depreciated_0",
    "depreciated_1",
    "depreciated_2",
    "depreciated_3",
    "depreciated_4",
    "unknown_2",
    "gametype",
    "unknown_3",
    "state_right",
    "state_left",
    "state_center",
    "unknown_4",
    "gametype_friendly",
    "gametype_enemy",
    "metagame_top",
    "metagame_player_1",
    "metagame_player_2",
    "metagame_player_3",
    "metagame_player_4",
    "theater",
    )

chdt_hud_widget_render_data_input = (
    "zero",
    "one",
    "time",
    "fade",
    "unit_shield_current",
    "unit_shield",
    "clip_ammo_fraction",
    "total_ammo_fraction",
    "heat_fraction",
    "battery_fraction",
    "pickup",
    "unit_autoaimed",
    "grenade",
    "grenade_fraction",
    "charge_fraction",
    "friendly_score",
    "enemy_score",
    "score_to_win",
    "arming_fraction",
    "unknown_0",
    "unit_1x_overshield_current",
    "unit_1x_overshield",
    "unit_2x_overshield_current",
    "unit_2x_overshield",
    "unit_3x_overshield_current",
    "unit_3x_overshield",
    "aim_yaw",
    "aim_pitch",
    "target_distance",
    "target_elevation",
    "editor_budget",
    "editor_budget_cost",
    "film_total_time",
    "film_current_time",
    "unknown_1",
    "film_timeline_fraction_1",
    "film_timeline_fraction_2",
    "unknown_2",
    "unknown_3",
    "metagame_time",
    "metagame_score_transient",
    "metagame_score_player_1",
    "metagame_score_player_2",
    "metagame_score_player_3",
    "metagame_score_player_4",
    "metagame_modifier",
    "unknown_4",
    "sensor_range",
    "netdebug_latency",
    "netdebug_latency_quality",
    "netdebug_host_quality",
    "netdebug_local_quality",
    "metagame_score_negative",
    )

chdt_hud_widget_render_data_output_color_a = (
    "local_a",
    "local_b",
    "local_c",
    "local_d",
    "unknown_4",
    "unknown_5",
    "scoreboard_friendly",
    "scoreboard_enemy",
    "arming_team",
    "metagame_player_1",
    "metagame_player_2",
    "metagame_player_3",
    "metagame_player_4",
    "unknown_13",
    "global_dynamic_0",
    "global_dynamic_1",
    "global_dynamic_2",
    "global_dynamic_3",
    "global_dynamic_4",
    "global_dynamic_5",
    "global_dynamic_6",
    "global_dynamic_7",
    "global_dynamic_8",
    "global_dynamic_9",
    "global_dynamic_10",
    "global_dynamic_11",
    "global_dynamic_12",
    "global_dynamic_13",
    "global_dynamic_14",
    "global_dynamic_15",
    "global_dynamic_16",
    "global_dynamic_17",
    "global_dynamic_18",
    "global_dynamic_19",
    "global_dynamic_20",
    "global_dynamic_21",
    "global_dynamic_22",
    "global_dynamic_23",
    "global_dynamic_24",
    "global_dynamic_25",
    "global_dynamic_26",
    "global_dynamic_27",
    )

chdt_hud_widget_render_data_output_scalar_a = (
    "input",
    "range_input",
    "local_a",
    "local_b",
    "local_c",
    "local_d",
    "unknown_6",
    "unknown_7",
    )

chdt_hud_widget_render_data_shader_index = (
    "simple",
    "meter",
    "text_simple",
    "meter_shield",
    "meter_gradient",
    "crosshair",
    "directional_damage",
    "solid",
    "sensor",
    "meter_single_color",
    "navpoint",
    "medal",
    "texture_cam",
    "cortana_screen",
    "cortana_camera",
    "cortana_offscreen",
    "cortana_screen_final",
    "meter_chapter",
    "meter_double_gradient",
    "meter_radial_gradient",
    "turbulence",
    "emblem",
    "cortana_composite",
    "directional_damage_apply",
    "really_simple",
    )

chdt_hud_widget_special_hud_type = (
    "unspecial",
    "ammo",
    "crosshair_and_scope",
    "unit_shield_meter",
    "grenades",
    "gametype",
    "motion_sensor",
    "spike_grenade",
    "firebomb_grenade",
    )

chdt_hud_widget_text_widget_font = (
    "conduit_18_0",
    "fixedsys_9_0",
    "fixedsys_9_1",
    "conduit_16_0",
    "conduit_32_0",
    "conduit_32_1",
    "conduit_23",
    "larabie_10",
    "conduit_18_1",
    "conduit_16_1",
    "pragmata_14",
    )


chdt_hud_widget_state_data = Struct("state_data",
    Bool16("_1_engine",
        ("capture_the_flag", 1 << 4),
        "slayer",
        "oddball",
        "king_of_the_hill",
        "juggernaut",
        "territories",
        "assault",
        "vip",
        "infection",
        ("editor", 1 << 14),
        "theater",
        ),
    Bool16("_2",
        "biped_1",
        "biped_2",
        "biped_3",
        ),
    Bool16("_3",
        "offense",
        "defense",
        "free_for_all",
        ("talking_disabled", 1 << 6),
        "tap_to_talk",
        "talking_enabled",
        "not_talking",
        "talking",
        ),
    Bool16("_4_resolution", *unknown_flags_16),
    Bool16("_5_scoreboard",
        "has_friends",
        "has_enemies",
        "has_variant_name",
        "someone_is_talking",
        "is_arming",
        "time_enabled",
        "friends_have_x",
        "enemies_have_x",
        "friends_are_x",
        "enemies_are_x",
        "x_is_down",
        "summary_enabled",
        "netdebug",
        ),
    Bool16("_6",
        "texture_cam_enabled",
        "autoaim",
        ("training_prompt", 1 << 4),
        "objective_prompt",
        ),
    Bool16("_7_editor",
        "editor_inactive",
        "editor_active",
        "editor_holding",
        "editor_not_allowed",
        "is_editor_biped",
        ),
    Bool16("_8",
        "motion_tracker_10m",
        "motion_tracker_25m",
        "motion_tracker_75m",
        "motion_tracker_150m",
        ("metagame_player_2_exists", 1 << 6),
        ("metagame_player_3_exists", 1 << 8),
        ("metagame_player_4_exists", 1 << 10),
        ("metagame_score_added", 1 << 12),
        ("metagame_score_removed", 1 << 14),
        ),
    Bool16("_9",
        "pickup_grenades",
        ),
    Bool16("_10",
        "binoculars_enabled",
        "unit_is_zoomed_level_1",
        "unit_is_zoomed_level_2",
        ),
    Bool16("_11",
        "primary_weapon",
        "secondary_weapon",
        ),
    Bool16("_12",
        "motion_tracker_enabled",
        ("selected_frag_grenades", 1 << 2),
        "selected_plasma_grenades",
        "selected_spike_grenades",
        "selected_fire_grenades",
        ("has_1x_overshield", 1 << 12),
        "has_2x_overshield",
        "has_3x_overshield",
        "has_shields",
        ),
    Bool16("_13",
        ("pickup_ammo", 1 << 1),
        ),
    Bool16("_14",
        "primary_weapon",
        "secondary_weapon",
        "backpack",
        ),
    Bool16("_15",
        "not_autoaim",
        "autoaim_friendly",
        "autoaim_enemy",
        "autoaim_headshot",
        ("plasma_locked_on", 1 << 7),
        ),
    Bool16("_16",
        ("missile_locked", 1 << 1),
        "missile_locking",
        ),
    Bool16("_17",
        ("has_frag_grenades", 1 << 2),
        "has_plasma_grenades",
        "has_spike_grenades",
        "has_fire_grenades",
        ),
    Bool16("_18_ammo",
        "clip_warning",
        "ammo_warning",
        ("low_battery_1", 1 << 4),
        "low_battery_2",
        "overheated",
        ),
    Bool16("_19",
        "binoculars_enabled",
        "unit_is_zoomed_level_1",
        "unit_is_zoomed_level_2",
        ),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=40
    )


chdt_hud_widget_placement_data = Struct("placement_data",
    SEnum16("anchor", *chdt_hud_widget_placement_data_anchor),
    SInt16("unknown", VISIBLE=False),
    QStruct("mirror_offset", INCLUDE=xy_float),
    QStruct("offset", INCLUDE=xy_float),
    QStruct("scale", INCLUDE=xy_float),
    ENDIAN=">", SIZE=28
    )


chdt_hud_widget_animation_data = Struct("animation_data",
    Bool16("animation_1_flags",
        "reverse_frames",
        ),
    SEnum16("animation_1_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_1"),
    Bool16("animation_2_flags",
        "reverse_frames",
        ),
    SEnum16("animation_2_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_2"),
    Bool16("animation_3_flags",
        "reverse_frames",
        ),
    SEnum16("animation_3_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_3"),
    Bool16("animation_4_flags",
        "reverse_frames",
        ),
    SEnum16("animation_4_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_4"),
    Bool16("animation_5_flags",
        "reverse_frames",
        ),
    SEnum16("animation_5_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_5"),
    Bool16("animation_6_flags",
        "reverse_frames",
        ),
    SEnum16("animation_6_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_6"),
    ENDIAN=">", SIZE=120
    )


chdt_hud_widget_render_data = Struct("render_data",
    SEnum16("shader_index", *chdt_hud_widget_render_data_shader_index),
    SInt16("unknown", VISIBLE=False),
    SEnum16("input", *chdt_hud_widget_render_data_input),
    SEnum16("range_input", *chdt_hud_widget_render_data_input),
    color_argb_uint32("local_color_a"),
    color_argb_uint32("local_color_b"),
    color_argb_uint32("local_color_c"),
    color_argb_uint32("local_color_d"),
    Float("local_scalar_a"),
    Float("local_scalar_b"),
    Float("local_scalar_c"),
    Float("local_scalar_d"),
    SEnum16("output_color_a", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_b", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_c", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_d", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_e", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_f", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_scalar_a", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_b", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_c", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_d", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_e", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_f", *chdt_hud_widget_render_data_output_scalar_a),
    ENDIAN=">", SIZE=64
    )


chdt_hud_widget_bitmap_widget_state_data = Struct("state_data",
    Bool16("_1_engine",
        ("capture_the_flag", 1 << 4),
        "slayer",
        "oddball",
        "king_of_the_hill",
        "juggernaut",
        "territories",
        "assault",
        "vip",
        "infection",
        ("editor", 1 << 14),
        "theater",
        ),
    Bool16("_2",
        "biped_1",
        "biped_2",
        "biped_3",
        ),
    Bool16("_3",
        "offense",
        "defense",
        "free_for_all",
        ("talking_disabled", 1 << 6),
        "tap_to_talk",
        "talking_enabled",
        "not_talking",
        "talking",
        ),
    Bool16("_4_resolution", *unknown_flags_16),
    Bool16("_5_scoreboard",
        "has_friends",
        "has_enemies",
        "has_variant_name",
        "someone_is_talking",
        "is_arming",
        "time_enabled",
        "friends_have_x",
        "enemies_have_x",
        "friends_are_x",
        "enemies_are_x",
        "x_is_down",
        "summary_enabled",
        "netdebug",
        ),
    Bool16("_6",
        "texture_cam_enabled",
        "autoaim",
        ("training_prompt", 1 << 4),
        "objective_prompt",
        ),
    Bool16("_7_editor",
        "editor_inactive",
        "editor_active",
        "editor_holding",
        "editor_not_allowed",
        "is_editor_biped",
        ),
    Bool16("_8",
        "motion_tracker_10m",
        "motion_tracker_25m",
        "motion_tracker_75m",
        "motion_tracker_150m",
        ("metagame_player_2_exists", 1 << 6),
        ("metagame_player_3_exists", 1 << 8),
        ("metagame_player_4_exists", 1 << 10),
        ("metagame_score_added", 1 << 12),
        ("metagame_score_removed", 1 << 14),
        ),
    Bool16("_9",
        "pickup_grenades",
        ),
    Bool16("_10",
        "binoculars_enabled",
        "unit_is_zoomed_level_1",
        "unit_is_zoomed_level_2",
        ),
    Bool16("_11",
        "primary_weapon",
        "secondary_weapon",
        ),
    Bool16("_12",
        "motion_tracker_enabled",
        ("selected_frag_grenades", 1 << 2),
        "selected_plasma_grenades",
        "selected_spike_grenades",
        "selected_fire_grenades",
        ("has_1x_overshield", 1 << 12),
        "has_2x_overshield",
        "has_3x_overshield",
        "has_shields",
        ),
    Bool16("_13",
        ("pickup_ammo", 1 << 1),
        ),
    Bool16("_14",
        "primary_weapon",
        "secondary_weapon",
        "backpack",
        ),
    Bool16("_15",
        "not_autoaim",
        "autoaim_friendly",
        "autoaim_enemy",
        "autoaim_headshot",
        ("plasma_locked_on", 1 << 7),
        ),
    Bool16("_16",
        ("missile_locked", 1 << 1),
        "missile_locking",
        ),
    Bool16("_17",
        ("has_frag_grenades", 1 << 2),
        "has_plasma_grenades",
        "has_spike_grenades",
        "has_fire_grenades",
        ),
    Bool16("_18_ammo",
        "clip_warning",
        "ammo_warning",
        ("low_battery_1", 1 << 4),
        "low_battery_2",
        "overheated",
        ),
    Bool16("_19",
        "binoculars_enabled",
        "unit_is_zoomed_level_1",
        "unit_is_zoomed_level_2",
        ),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=40
    )


chdt_hud_widget_bitmap_widget_placement_data = Struct("placement_data",
    SEnum16("anchor", *chdt_hud_widget_placement_data_anchor),
    SInt16("unknown", VISIBLE=False),
    QStruct("mirror_offset", INCLUDE=xy_float),
    QStruct("offset", INCLUDE=xy_float),
    QStruct("scale", INCLUDE=xy_float),
    ENDIAN=">", SIZE=28
    )


chdt_hud_widget_bitmap_widget_animation_data = Struct("animation_data",
    Bool16("animation_1_flags",
        "reverse_frames",
        ),
    SEnum16("animation_1_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_1"),
    Bool16("animation_2_flags",
        "reverse_frames",
        ),
    SEnum16("animation_2_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_2"),
    Bool16("animation_3_flags",
        "reverse_frames",
        ),
    SEnum16("animation_3_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_3"),
    Bool16("animation_4_flags",
        "reverse_frames",
        ),
    SEnum16("animation_4_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_4"),
    Bool16("animation_5_flags",
        "reverse_frames",
        ),
    SEnum16("animation_5_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_5"),
    Bool16("animation_6_flags",
        "reverse_frames",
        ),
    SEnum16("animation_6_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_6"),
    ENDIAN=">", SIZE=120
    )


chdt_hud_widget_bitmap_widget_render_data = Struct("render_data",
    SEnum16("shader_index", *chdt_hud_widget_render_data_shader_index),
    SInt16("unknown", VISIBLE=False),
    SEnum16("input", *chdt_hud_widget_render_data_input),
    SEnum16("range_input", *chdt_hud_widget_render_data_input),
    color_argb_uint32("local_color_a"),
    color_argb_uint32("local_color_b"),
    color_argb_uint32("local_color_c"),
    color_argb_uint32("local_color_d"),
    Float("local_scalar_a"),
    Float("local_scalar_b"),
    Float("local_scalar_c"),
    Float("local_scalar_d"),
    SEnum16("output_color_a", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_b", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_c", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_d", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_e", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_f", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_scalar_a", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_b", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_c", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_d", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_e", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_f", *chdt_hud_widget_render_data_output_scalar_a),
    ENDIAN=">", SIZE=64
    )


chdt_hud_widget_bitmap_widget = Struct("bitmap_widget",
    h3_string_id("name"),
    SEnum16("special_hud_type", *chdt_hud_widget_special_hud_type),
    UInt8("unknown_0", VISIBLE=False),
    UInt8("unknown_1", VISIBLE=False),
    h3_reflexive("state_data", chdt_hud_widget_bitmap_widget_state_data),
    h3_reflexive("placement_data", chdt_hud_widget_bitmap_widget_placement_data),
    h3_reflexive("animation_data", chdt_hud_widget_bitmap_widget_animation_data),
    h3_reflexive("render_data", chdt_hud_widget_bitmap_widget_render_data),
    SInt32("widget_index"),
    Bool16("flags",
        "mirror_horizontally",
        "mirror_vertically",
        "stretch_edges",
        "enable_texture_cam",
        "looping",
        ("player_1_emblem", 1 << 6),
        "player_2_emblem",
        "player_3_emblem",
        "player_4_emblem",
        ),
    SInt16("unknown_2", VISIBLE=False),
    h3_dependency("bitmap"),
    UInt8("bitmap_sprite_index"),
    UInt8("unknown_3", VISIBLE=False),
    UInt8("unknown_4", VISIBLE=False),
    UInt8("unknown_5", VISIBLE=False),
    ENDIAN=">", SIZE=84
    )


chdt_hud_widget_text_widget_state_data = Struct("state_data",
    Bool16("_1_engine",
        ("capture_the_flag", 1 << 4),
        "slayer",
        "oddball",
        "king_of_the_hill",
        "juggernaut",
        "territories",
        "assault",
        "vip",
        "infection",
        ("editor", 1 << 14),
        "theater",
        ),
    Bool16("_2",
        "biped_1",
        "biped_2",
        "biped_3",
        ),
    Bool16("_3",
        "offense",
        "defense",
        "free_for_all",
        ("talking_disabled", 1 << 6),
        "tap_to_talk",
        "talking_enabled",
        "not_talking",
        "talking",
        ),
    Bool16("_4_resolution", *unknown_flags_16),
    Bool16("_5_scoreboard",
        "has_friends",
        "has_enemies",
        "has_variant_name",
        "someone_is_talking",
        "is_arming",
        "time_enabled",
        "friends_have_x",
        "enemies_have_x",
        "friends_are_x",
        "enemies_are_x",
        "x_is_down",
        "summary_enabled",
        "netdebug",
        ),
    Bool16("_6",
        "texture_cam_enabled",
        "autoaim",
        ("training_prompt", 1 << 4),
        "objective_prompt",
        ),
    Bool16("_7_editor",
        "editor_inactive",
        "editor_active",
        "editor_holding",
        "editor_not_allowed",
        "is_editor_biped",
        ),
    Bool16("_8",
        "motion_tracker_10m",
        "motion_tracker_25m",
        "motion_tracker_75m",
        "motion_tracker_150m",
        ("metagame_player_2_exists", 1 << 6),
        ("metagame_player_3_exists", 1 << 8),
        ("metagame_player_4_exists", 1 << 10),
        ("metagame_score_added", 1 << 12),
        ("metagame_score_removed", 1 << 14),
        ),
    Bool16("_9",
        "pickup_grenades",
        ),
    Bool16("_10",
        "binoculars_enabled",
        "unit_is_zoomed_level_1",
        "unit_is_zoomed_level_2",
        ),
    Bool16("_11",
        "primary_weapon",
        "secondary_weapon",
        ),
    Bool16("_12",
        "motion_tracker_enabled",
        ("selected_frag_grenades", 1 << 2),
        "selected_plasma_grenades",
        "selected_spike_grenades",
        "selected_fire_grenades",
        ("has_1x_overshield", 1 << 12),
        "has_2x_overshield",
        "has_3x_overshield",
        "has_shields",
        ),
    Bool16("_13",
        ("pickup_ammo", 1 << 1),
        ),
    Bool16("_14",
        "primary_weapon",
        "secondary_weapon",
        "backpack",
        ),
    Bool16("_15",
        "not_autoaim",
        "autoaim_friendly",
        "autoaim_enemy",
        "autoaim_headshot",
        ("plasma_locked_on", 1 << 7),
        ),
    Bool16("_16",
        ("missile_locked", 1 << 1),
        "missile_locking",
        ),
    Bool16("_17",
        ("has_frag_grenades", 1 << 2),
        "has_plasma_grenades",
        "has_spike_grenades",
        "has_fire_grenades",
        ),
    Bool16("_18_ammo",
        "clip_warning",
        "ammo_warning",
        ("low_battery_1", 1 << 4),
        "low_battery_2",
        "overheated",
        ),
    Bool16("_19",
        "binoculars_enabled",
        "unit_is_zoomed_level_1",
        "unit_is_zoomed_level_2",
        ),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=40
    )


chdt_hud_widget_text_widget_placement_data = Struct("placement_data",
    SEnum16("anchor", *chdt_hud_widget_placement_data_anchor),
    SInt16("unknown", VISIBLE=False),
    QStruct("mirror_offset", INCLUDE=xy_float),
    QStruct("offset", INCLUDE=xy_float),
    QStruct("scale", INCLUDE=xy_float),
    ENDIAN=">", SIZE=28
    )


chdt_hud_widget_text_widget_animation_data = Struct("animation_data",
    Bool16("animation_1_flags",
        "reverse_frames",
        ),
    SEnum16("animation_1_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_1"),
    Bool16("animation_2_flags",
        "reverse_frames",
        ),
    SEnum16("animation_2_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_2"),
    Bool16("animation_3_flags",
        "reverse_frames",
        ),
    SEnum16("animation_3_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_3"),
    Bool16("animation_4_flags",
        "reverse_frames",
        ),
    SEnum16("animation_4_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_4"),
    Bool16("animation_5_flags",
        "reverse_frames",
        ),
    SEnum16("animation_5_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_5"),
    Bool16("animation_6_flags",
        "reverse_frames",
        ),
    SEnum16("animation_6_function", *chdt_hud_widget_animation_data_animation_1_function),
    h3_dependency("animation_6"),
    ENDIAN=">", SIZE=120
    )


chdt_hud_widget_text_widget_render_data = Struct("render_data",
    SEnum16("shader_index", *chdt_hud_widget_render_data_shader_index),
    SInt16("unknown", VISIBLE=False),
    SEnum16("input", *chdt_hud_widget_render_data_input),
    SEnum16("range_input", *chdt_hud_widget_render_data_input),
    color_argb_uint32("local_color_a"),
    color_argb_uint32("local_color_b"),
    color_argb_uint32("local_color_c"),
    color_argb_uint32("local_color_d"),
    Float("local_scalar_a"),
    Float("local_scalar_b"),
    Float("local_scalar_c"),
    Float("local_scalar_d"),
    SEnum16("output_color_a", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_b", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_c", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_d", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_e", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_color_f", *chdt_hud_widget_render_data_output_color_a),
    SEnum16("output_scalar_a", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_b", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_c", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_d", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_e", *chdt_hud_widget_render_data_output_scalar_a),
    SEnum16("output_scalar_f", *chdt_hud_widget_render_data_output_scalar_a),
    ENDIAN=">", SIZE=64
    )


chdt_hud_widget_text_widget = Struct("text_widget",
    h3_string_id("name"),
    SEnum16("special_hud_type", *chdt_hud_widget_special_hud_type),
    UInt8("unknown_0", VISIBLE=False),
    UInt8("unknown_1", VISIBLE=False),
    h3_reflexive("state_data", chdt_hud_widget_text_widget_state_data),
    h3_reflexive("placement_data", chdt_hud_widget_text_widget_placement_data),
    h3_reflexive("animation_data", chdt_hud_widget_text_widget_animation_data),
    h3_reflexive("render_data", chdt_hud_widget_text_widget_render_data),
    SInt32("widget_index"),
    Bool16("flags",
        "string_is_a_number",
        "force_2_digit",
        "force_3_digit",
        "prefix_0",
        "m_suffix",
        "hundredths_decimal",
        "thousandths_decimal",
        "hundred_thousandths_decimal",
        "only_a_number",
        "x_suffix",
        "in_brackets",
        "time_format_s_ms",
        "time_format_h_m_s",
        "money_format",
        "prefix_1",
        ),
    SEnum16("font", *chdt_hud_widget_text_widget_font),
    h3_string_id("string"),
    ENDIAN=">", SIZE=68
    )


chdt_hud_widget = Struct("hud_widget",
    h3_string_id("name"),
    SEnum16("special_hud_type", *chdt_hud_widget_special_hud_type),
    UInt8("unknown_0", VISIBLE=False),
    UInt8("unknown_1", VISIBLE=False),
    h3_reflexive("state_data", chdt_hud_widget_state_data),
    h3_reflexive("placement_data", chdt_hud_widget_placement_data),
    h3_reflexive("animation_data", chdt_hud_widget_animation_data),
    h3_reflexive("render_data", chdt_hud_widget_render_data),
    h3_reflexive("bitmap_widgets", chdt_hud_widget_bitmap_widget),
    h3_reflexive("text_widgets", chdt_hud_widget_text_widget),
    ENDIAN=">", SIZE=80
    )


chdt_body = Struct("tagdata",
    h3_reflexive("hud_widgets", chdt_hud_widget),
    SInt32("low_clip_cutoff"),
    SInt32("low_ammo_cutoff"),
    SInt32("age_cutoff"),
    ENDIAN=">", SIZE=24
    )


def get():
    return chdt_def

chdt_def = TagDef("chdt",
    h3_blam_header('chdt'),
    chdt_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["chdt"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

def get(): return font_def


character_table = QStruct("character_table",
    SInt16("character_index"),
    SIZE=2
    )

character_tables = Struct("character_tables",
    reflexive("character_table", character_table, 256),
    SIZE=12
    )

character = QStruct("character",
    UInt16("character"),
    SInt16("character_width"),
    SInt16("bitmap_width", EDITABLE=False),
    SInt16("bitmap_height", EDITABLE=False),
    SInt16("bitmap_origin_x"),
    SInt16("bitmap_origin_y"),
    SInt16("hardware_character_index"),
    Pad(2),
    SInt32("pixels_offset", EDITABLE=False),
    SIZE=20, WIDGET=FontCharacterFrame
    )

font_body = Struct("tagdata",
    SInt32("flags"),
    SInt16("ascending_height"),
    SInt16("decending_height"),
    SInt16("leading_height"),
    SInt16("leading_width"),
    Pad(36),

    reflexive("character_tables", character_tables, 256),
    dependency("bold", "font"),
    dependency("italic", "font"),
    dependency("condense", "font"),
    dependency("underline", "font"),
    reflexive("characters", character, 65535),
    rawdata_ref("pixels", max_size=8388608),
    SIZE=156,
    )

font_def = TagDef("font",
    blam_header('font'),
    font_body,

    ext=".font", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

material = Struct("material",
    dependency("effect", "effe"),
    dependency("sound", "snd!"),
    SIZE=48,
    )

effect = Struct("effect",
    reflexive("materials", material, len(materials_list), *materials_list),
    SIZE=28,
    )

foot_body = Struct("tagdata",
    reflexive("effects", effect, 13, *material_effect_types),
    SIZE=140,
    )



def get():
    return foot_def

foot_def = TagDef("foot",
    blam_header('foot'),
    foot_body,

    ext=".material_effects", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

item_permutation = Struct("permutation",
    Pad(32),
    Float("weight"),
    dependency("item", valid_items),
    SIZE=84,
    )

itmc_body = Struct("tagdata",
    reflexive("item_permutations", item_permutation, 32767,
        DYN_NAME_PATH='.item.filepath'),
    SInt16("spawn_time", SIDETIP="seconds(0 = default)",
            UNIT_SCALE=per_sec_unit_scale),
    SIZE=92,
    )


def get():
    return itmc_def

itmc_def = TagDef("itmc",
    blam_header('itmc',0),
    itmc_body,

    ext=".item_collection", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.phys import PhysTag
from supyr_struct.defs.tag_def import TagDef

inertial_matrix = Struct("inertial_matrix",
    QStruct("yy_zz", GUI_NAME="yy+zz    -xy    -zx", INCLUDE=ijk_float),
    QStruct("zz_xx", GUI_NAME="-xy    zz+xx    -yz", INCLUDE=ijk_float),
    QStruct("xx_yy", GUI_NAME="-zx      -yz  xx+yy", INCLUDE=ijk_float),
    SIZE=36,
    )

powered_mass_point = Struct("powered_mass_point",
    ascii_str32("name"),
    Bool32('flags',
        'ground_friction',
        'water_friction',
        'air_friction',
        'water_lift',
        'air_lift',
        'thrust',
        'antigrav',
        ),
    Float("antigrav_strength"),
    Float("antigrav_offset"),
    Float("antigrav_height"),
    Float("antigrav_damp_fraction"),
    Float("antigrav_normal_k1"),
    Float("antigrav_normal_k0"),
    SIZE=128,
    )

mass_point = Struct("mass_point",
    ascii_str32("name"),
    dyn_senum16("powered_mass_point",
        DYN_NAME_PATH="tagdata.powered_mass_points.STEPTREE[DYN_I].name"),
    SInt16("model_node"),
    Bool32('flags',
        'metallic',
        ),
    Float("relative_mass", MIN=0.0),
    Float("mass", VISIBLE=False, MIN=0.0),
    Float("relative_density", MIN=0.0),
    Float("density", VISIBLE=False, MIN=0.0),
    QStruct("position", INCLUDE=xyz_float),
    QStruct("forward", INCLUDE=ijk_float),
    QStruct("up", INCLUDE=ijk_float),
    SEnum16('friction_type',
        'point',
        'forward',
        'left',
        'up',
        ),
    Pad(2),
    Float("friction_parallel_scale"),
    Float("friction_perpendicular_scale"),
    Float("radius", MIN=0.0, ALLOW_MIN=False),
    SIZE=128,
    )

phys_body = Struct("tagdata",
    Float("radius"),
    Float("moment_scale"),
    Float("mass", MIN=0.0),
    QStruct("center_of_mass", INCLUDE=xyz_float, VISIBLE=False),
    Float("density", MIN=0.0),
    Float("gravity_scale"),
    Float("ground_friction", UNIT_SCALE=per_sec_unit_scale),
    Float("ground_depth"),
    Float("ground_damp_fraction"),
    Float("ground_normal_k1"),
    Float("ground_normal_k0"),
    Pad(4),
    Float("water_friction", UNIT_SCALE=per_sec_unit_scale),
    Float("water_depth"),
    Float("water_density"),
    Pad(4),
    Float("air_friction", UNIT_SCALE=per_sec_unit_scale),
    Pad(4),
    Float("xx_moment", VISIBLE=False),
    Float("yy_moment", VISIBLE=False),
    Float("zz_moment", VISIBLE=False),

    reflexive("inertia_matrices", inertial_matrix, 2,
        "regular", "inverse", MIN=2, MAX=2, VISIBLE=False),
    reflexive("powered_mass_points", powered_mass_point, 32,
        DYN_NAME_PATH='.name'),
    reflexive("mass_points", mass_point, 32,
        DYN_NAME_PATH='.name'),
    SIZE=128,
    COMMENT="\
Some fields have been hidden because you cant edit them.\n\n\
This is because they are recalculated when you hit save.\n\
The inertial matrices and xx/yy/zz moments are hidden as\n\
well as the mass and density of the individual mass points."
    )


def get():
    return phys_def

phys_def = TagDef("phys",
    blam_header('phys', 4),
    phys_body,

    ext=".physics", endian=">", tag_cls=PhysTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.hek.defs.objs.bitm import BitmTag
from reclaimer.bitmaps.p8_palette import STUBBS_P8_PALETTE

class StubbsBitmTag(BitmTag):
    def __init__(self, *args, **kwargs):
        BitmTag.__init__(self, *args, **kwargs)
        self.p8_palette = STUBBS_P8_PALETTE



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.field_types import *
from reclaimer.h2.field_type_methods import h2_rawdata_ref_parser,\
     h2_tag_ref_parser

H2TagRef = FieldType(base=TagRef, name="H2TagRef", parser=h2_tag_ref_parser)
H2Reflexive  = FieldType(base=Reflexive,   name="H2Reflexive")
H2RawdataRef = FieldType(base=RawdataRef,  name="H2RawdataRef",
                         parser=h2_rawdata_ref_parser)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from struct import pack, unpack
from types import MethodType

from reclaimer.field_types import *
from reclaimer.constants import *
from reclaimer.common_descs import ascii_str32,\
     script_types as h1_script_types,\
     script_object_types as h1_script_object_types
from reclaimer.util import float_to_str
from reclaimer.h2.common_descs import script_types as h2_script_types,\
     script_object_types as h2_script_object_types

from supyr_struct.defs.block_def import BlockDef
from supyr_struct.field_types import FieldType

try:
    from reclaimer.enums import _script_built_in_functions_test
except Exception:
    _script_built_in_functions_test = None

# in a list that begins with the keyed expression, this
# is the number of nodes required before we will indent them.
INDENT_LIST_MIN = {
    "begin": 2, "begin_random": 2, "cond": 2, "if": 3,
    "+": 4, "*": 4, "and": 4, "or": 4, "min": 4, "max": 4,
    "ai_debug_communication_suppress": 4,
    "ai_debug_communication_ignore": 4,
    "ai_debug_communication_focus": 4,
    }

# in a list that begins with the keyed expression, this
# is the node index we will start applying indentation at.
START_INDENTING_AFTER = {
    "begin": 1, "begin_random": 1, "cond": 1, "if": 2,
    "+": 1, "*": 1, "and": 1, "or": 1, "min": 1, "max": 1,
    "ai_debug_communication_suppress": 1,
    "ai_debug_communication_ignore": 1,
    "ai_debug_communication_focus": 1,
    }

HSC_IS_PRIMITIVE   = 1 << 0
HSC_IS_SCRIPT_CALL = 1 << 1
HSC_IS_GLOBAL      = 1 << 2
HSC_IS_GARBAGE_COLLECTABLE = 1 << 3
HSC_IS_SCRIPT_OR_GLOBAL = HSC_IS_SCRIPT_CALL | HSC_IS_GLOBAL

SCRIPT_OBJECT_TYPES_TO_SCENARIO_REFLEXIVES = dict((
    (10, "scripts"), (11, "trigger_volumes"), (12, "cutscene_flags"),
    (13, "cutscene_camera_points"), (14, "cutscene_titles"),
    (15, "recorded_animations"), (16, "device_groups"),
    (17, "encounters"), (18, "command_lists"),
    (19, "player_starting_profiles"), (20, "ai_conversations"),
    ) + tuple((i, "object_names") for i in range(37, 49)))


h1_script_type = SEnum16("type", *h1_script_types)
h1_return_type = SEnum16("return_type", *h1_script_object_types)


# this is a more complete version of the fast_script_node def below
script_node = Struct("script_node",
    UInt16("salt"),
    Union("index_union",
        CASES={
            "constant_type":  Struct("constant_type", SInt16("value")),
            "function_index": Struct("function_index", SInt16("value")),
            "script_index":   Struct("script_index", SInt16("value")),
            },
        COMMENT="""
For most intents and purposes, this value mirrors the 'type' field"""
        ),
    SEnum16("type", *h1_script_object_types),
    Bool16("flags",
        "is_primitive",
        "is_script_call",
        "is_global",
        "is_garbage_collectable",
        ),
    UInt32("next_node"),
    UInt32("string_offset"),
    Union("data",
        CASES={
            "bool":  Struct("bool", UInt8("data")),
            "int16": Struct("int16", SInt16("data")),
            "int32": Struct("int32", SInt32("data")),
            "real":  Struct("real", Float("data")),
            "node":  Struct("node", UInt32("data")),
            }
        ),
    SIZE=20
    )

fast_script_node = QStruct("script_node",
    UInt16("salt"),
    UInt16("index_union"),
    UInt16("type"),
    UInt16("flags"),
    UInt32("next_node"),
    UInt32("string_offset"),
    UInt32("data"),
    SIZE=20
    )

h1_script_syntax_data = Struct("script syntax data header",
    ascii_str32('name', DEFAULT="script node"),
    UInt16("max_nodes", DEFAULT=19001),  # this is 1 more than expected
    UInt16("node_size", DEFAULT=20),
    UInt8("is_valid", DEFAULT=1),
    UInt8("identifier_zero_invalid"),   # zero?
    UInt16("unused"),
    UInt32("sig", DEFAULT="d@t@"),
    UInt16("next_node"),  # zero?
    UInt16("last_node"),
    BytesRaw("next", SIZE=4),
    Pointer32("first"),
    SIZE=56,
    STEPTREE=WhileArray("nodes", SUB_STRUCT=fast_script_node)
    )

h1_script_syntax_data_os = dict(h1_script_syntax_data)
h1_script_syntax_data_os[1] = UInt16("max_nodes", DEFAULT=28501)

h1_script_syntax_data_def    = BlockDef(h1_script_syntax_data)
h1_script_syntax_data_os_def = BlockDef(h1_script_syntax_data_os)


def cast_uint32_to_float(uint32, packer=MethodType(pack, "<I"),
                         unpacker=MethodType(unpack, "<f")):
    return unpacker(packer(uint32))[0]


def cast_uint32_to_sint16(uint32):
    return ((uint32 + 0x8000) % 0x10000) - 0x8000


def cast_uint32_to_sint32(uint32):
    return ((uint32 + 0x80000000) % 0x100000000) - 0x80000000


def get_hsc_node_string(string_data, node, hsc_node_strings_by_type=()):
    # if this is not a script or global, try to get the
    # string from the provided hsc_node_strings_by_type
    if (not(node.flags & HSC_IS_SCRIPT_OR_GLOBAL) and
            node.type in hsc_node_strings_by_type):
        hsc_node_strings = hsc_node_strings_by_type[node.type]

        # "ai" script object types(17) use 32 bits of the
        # data field to specify index instead of just 16.
        mask = 0xFFffFFff if node.type == 17 else 0xFFff
        if node.data & mask in hsc_node_strings:
            return hsc_node_strings[node.data & mask]

    end = string_data.find("\x00", node.string_offset)
    string = string_data[node.string_offset: end]
    #if _script_built_in_functions_test is not None and node.type == 2:
    #    if node.index_union not in range(len(_script_built_in_functions_test)):
    #        _script_built_in_functions_test.extend(
    #            [None] * (node.index_union + 1 -
    #                      len(_script_built_in_functions_test)))
    #    if _script_built_in_functions_test[node.index_union] is None:
    #        _script_built_in_functions_test[node.index_union] = string

    return string


def get_hsc_data_block(raw_syntax_data=None, engine="halo1"):
    block_def = None
    header_len = 56
    sig_off = 40

    if "yelo" in engine:
        block_def = h1_script_syntax_data_os_def
    elif "halo1" in engine:
        block_def = h1_script_syntax_data_def

    if block_def is None:
        return

    endianness_force = FieldType.force_little
    if raw_syntax_data[sig_off: sig_off+4] == b"d@t@":
        endianness_force = FieldType.force_big

    with endianness_force:
        syntax_data = block_def.build(rawdata=raw_syntax_data)
        node_size = syntax_data.node_size
        node_ct = min(syntax_data.max_nodes, syntax_data.last_node,
                      (len(raw_syntax_data) - header_len) // node_size)

        nodes = syntax_data.nodes
        for i in range(header_len, header_len + node_ct*node_size, node_size):
            nodes.append(rawdata=raw_syntax_data, root_offset=i)

    return syntax_data


def get_h1_scenario_script_object_type_strings(scnr_data):
    # NOTE: Still need to handle "hud_message", # 22
    script_strings_by_type = {}
    for script_object_type, reflexive_name in \
            SCRIPT_OBJECT_TYPES_TO_SCENARIO_REFLEXIVES.items():
        names = {}
        script_strings_by_type[script_object_type] = names

        i = 0
        for b in scnr_data[reflexive_name].STEPTREE:
            names[i] = b.name
            i += 1

    i = 0
    script_strings_by_type[35] = names = {}
    for b in scnr_data.bipeds_palette.STEPTREE:
        names[i] = b.name.filepath.split("/")[-1].split("\\")[-1]
        i += 1

    i = 0
    script_strings_by_type[17] = names = {}
    for enc in scnr_data.encounters.STEPTREE:
        j = 0
        for squad in enc.squads.STEPTREE:
            names[i + (j << 16) + 0x80000000] = "%s/%s" % (enc.name, squad.name)
            j += 1
        i += 1

    return script_strings_by_type


def get_node_sibling_count(node, nodes):
    ct = 0
    seen = set()
    while True:
        sib_index, salt = node.next_node & 0xFFff, node.next_node >> 16
        if salt == 0 or sib_index >= len(nodes):
            return ct
        elif sib_index in seen:
            return 0xFFffFFff

        ct += 1
        seen.add(sib_index)
        node = nodes[sib_index]


def get_first_significant_node(node, nodes, string_data, parent=None,
                               last_begin_parenthese=None, started=False):
    if (not started) or get_node_sibling_count(node, nodes) <= 1:
        # node has one or no siblings, meaning it COULD be a begin block
        if node.flags == HSC_IS_GARBAGE_COLLECTABLE:
            index, salt = node.data & 0xFFff, node.data >> 16
            if salt and index < len(nodes):
                # node is parenthese. go deeper
                return get_first_significant_node(
                    nodes[index], nodes, string_data, node, node, True)

        func_name = get_hsc_node_string(string_data, node)
        if ((node.flags & HSC_IS_PRIMITIVE) and
                node.type == 2 and func_name == "begin"):
            index, salt = node.next_node & 0xFFff, node.next_node >> 16
            if salt and index < len(nodes):
                return get_first_significant_node(
                    nodes[index], nodes, string_data, parent,
                    last_begin_parenthese, True)

    if last_begin_parenthese is not None:
        return last_begin_parenthese
    elif parent is not None:
        return parent
    return node


def decompile_node_bytecode(node_index, nodes, script_blocks, string_data,
                            object_types, indent=1, indent_size=4, **kwargs):
    if node_index < 0 or node_index >= len(nodes):
        return "", False, 0

    hsc_node_strings_by_type = kwargs.get("hsc_node_strings_by_type", ())
    has_newlines = False
    node_strs = []
    i = 0
    while node_index > 0:
        # collect unparsed strings of all linked nodes
        if node_index == 0xFFff:
            break

        newl = False
        node = nodes[node_index]
        node_type = node.type
        union_i = node.index_union
        node_str = ""
        if node.flags == HSC_IS_GARBAGE_COLLECTABLE:
            start_node = get_first_significant_node(node, nodes, string_data)

            child_node, salt = start_node.data & 0xFFff, start_node.data >> 16
            if salt != 0:
                node_str, newl, ct = decompile_node_bytecode(
                    child_node, nodes, script_blocks, string_data,
                    object_types, indent + 1, indent_size, **kwargs)

                if ct > 1 or newl or (node_str[:1]  != "(" and
                                      node_str[-1:] != ")"):
                    # only add a start parenthese so the end can be added
                    # on later when we decide how to pad the list elements
                    node_str = "(%s" % node_str[1:]
                else:
                    # reparse the node at a lesser indent
                    node_str, newl, ct = decompile_node_bytecode(
                        child_node, nodes, script_blocks, string_data,
                        object_types, indent, indent_size, **kwargs)

        elif node.flags & HSC_IS_GLOBAL:
            node_str = get_hsc_node_string(string_data, node,
                                           hsc_node_strings_by_type)
            if "global_uses" in kwargs:
                kwargs["global_uses"].add(node_str)

        elif (node.flags & HSC_IS_SCRIPT_CALL and
              union_i >= 0 and union_i < len(script_blocks)):
            # is_script_call is set
            block = script_blocks[union_i]
            node_str = "(%s" % block.name
            if "static_calls" in kwargs:
                kwargs["static_calls"].add(block.name)

        elif node.flags & HSC_IS_PRIMITIVE:
            # is_primitive is set
            if node_type in (2, 10):
                # function/script name
                node_str = get_hsc_node_string(string_data, node,
                                               hsc_node_strings_by_type)
                if node_type == 10 and "static_calls" in kwargs:
                    kwargs["static_calls"].add(node_str)
            elif node_type in (3, 4):
                # passthrough/void type
                pass
            elif node_type == 5:
                # bool
                node_str = "true" if node.data&1 else "false"
            elif node_type == 6:
                # float
                node_str = float_to_str(cast_uint32_to_float(node.data))
            elif node_type == 7:
                # short
                node_str = str(cast_uint32_to_sint16(node.data))
            elif node_type == 8:
                # long
                node_str = str(cast_uint32_to_sint32(node.data))
            elif node_type < len(object_types):
                # other
                node_str = '"%s"' % get_hsc_node_string(
                    string_data, node, hsc_node_strings_by_type)

        node_strs.append((node_str, newl))
        node_index, salt = node.next_node & 0xFFff, node.next_node >> 16
        if salt == 0:
            break

        i += 1

    node_str_ct = len(node_strs)
    if not node_str_ct:
        return "", False, 0

    string = ""
    indent_str = " " * indent_size * indent
    returned = False
    i = 0
    first_node = node_strs[0]
    node_indent_min   = INDENT_LIST_MIN.get(first_node[0], 20)
    node_indent_start = START_INDENTING_AFTER.get(first_node[0], 1)

    for node_str, local_newlines in node_strs:
        if not node_str:
            continue

        i += 1
        cap_end   = node_str[0] == "("
        do_indent = (i > node_indent_start and indent_str and
                     node_str_ct >= node_indent_min)
        has_newlines |= local_newlines

        # might need to indent this node's string
        if do_indent:
            if returned:
                node_str = indent_str + node_str
            else:
                node_str = "\n" + indent_str + node_str
                has_newlines = True
        elif i > 1:
            node_str = " " + node_str

        if cap_end:
            if local_newlines:
                node_str += "\n" + indent_str
            node_str += ")"

        returned = node_str[-1] == "\n"
        string += node_str

    if string:
        # always add a space to the beginning.
        # it can be stripped off by whatever we return to
        string = " " + string

    return string, has_newlines, i


def hsc_bytecode_to_string(syntax_data, string_data, block_index,
                           script_blocks, global_blocks, block_type,
                           engine="halo1", indent_size=4, **kwargs):
    if block_type not in ("script", "global"):
        return ""

    if ("halo1" in engine or "yelo" in engine or
        "stubbs" in engine or "shadowrun" in engine):
        script_types = h1_script_types
        object_types = h1_script_object_types
    elif "halo2" in engine:
        script_types = h2_script_types
        object_types = h2_script_object_types
    else:
        return ""

    if block_type == "global":
        block       = global_blocks[block_index]
        script_type = ""
        node_index  = block.initialization_expression_index & 0xFFFF
        main_type   = object_types[block.type.data]
    else:
        block       = script_blocks[block_index]
        script_type = script_types[block.type.data]
        node_index  = block.root_expression_index & 0xFFFF
        main_type   = object_types[block.return_type.data]
        if script_type in ("dormant", "startup", "continuous"):
            # these types wont compile if a return type is specified
            main_type = ""
        else:
            script_type += " "

    if "INVALID" in main_type:
        return ""

    indent_str  = " " * indent_size
    head = "%s %s%s %s" % (block_type, script_type, main_type, block.name)
    body, _, __ = decompile_node_bytecode(
        node_index, syntax_data.nodes, script_blocks, string_data,
        object_types, 1, indent_size, **kwargs)
    if block_type == "global":
        return "(%s%s)" % (head, body)
    return "(%s\n%s%s\n)" % (head, indent_str, body[1:])



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

def get():
    return srac_def

srac_def = TagDef("srac",
    blam_header('srac'),
    Struct('tagdata',
        ),

    ext=".shadowrun_actor", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.meta.halo1_map import tag_path_pointer, tag_index_array_pointer
from reclaimer.shadowrun_prototype.common_descs import *


sr_tag_header = Struct("tag header",
    UEnum32("class 1", GUI_NAME="primary tag class",   INCLUDE=sr_valid_tags),
    UEnum32("class 2", GUI_NAME="secondary tag class", INCLUDE=sr_valid_tags),
    UEnum32("class 3", GUI_NAME="tertiary tag class",  INCLUDE=sr_valid_tags),
    UInt32("id"),
    UInt32("path offset"),
    UInt32("meta offset"),
    UInt32("indexed"),
    # if indexed is 1, the meta_offset is the literal index in the
    # bitmaps, sounds, or loc cache that the meta data is located in.
    Pad(4),
    STEPTREE=CStrTagRef("path", POINTER=tag_path_pointer, MAX=768),
    SIZE=32
    )

sr_tag_index_array = TagIndex("tag index",
    SIZE=".tag_count", SUB_STRUCT=sr_tag_header,
    POINTER=tag_index_array_pointer
    )

sr_tag_index = Struct("tag index",
    UInt32("tag index offset"),
    UInt32("scenario tag id"),
    UInt32("map id"),  # normally unused, but the scenario tag's header
    #                    can be used for spoofing the maps checksum
    UInt32("tag count"),

    UInt32("vertex parts count"),
    UInt32("model data offset"),

    UInt32("index parts count"),
    UInt32("index parts offset"),
    UInt32("tag sig", EDITABLE=False, DEFAULT='tags'),

    SIZE=36,
    STEPTREE=sr_tag_index_array
    )

sr_tag_index_def  = BlockDef(sr_tag_index)



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path

from reclaimer.hek.handler import HaloHandler
from reclaimer.os_hek.defs import __all__ as all_def_names


class OsHaloHandler(HaloHandler):
    frozen_imp_paths = all_def_names
    default_defs_path = "reclaimer.os_hek.defs"



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import math

from reclaimer.model.jms import JmsNode, JmsMaterial, JmsVertex,\
     JmsTriangle, JmsModel
from traceback import format_exc


def jms_model_from_obj(obj_string, model_name=None):
    if not isinstance(obj_string, str):
        raise TypeError(
            "Argument must be of type 'str', not '%s'." % type(obj_string))

    if model_name is None:
        model_name = "__unnamed"

    model = JmsModel(model_name)
    model.nodes = [JmsNode("frame")]
    model.regions = ["__unnamed"]

    mats = model.materials
    tris = model.tris

    f_num = 0
    mat_num = 0
    mat_names = {}
    v_locs = []
    v_uvs = []
    v_norms = []
    v_datas = {}
    v_data_by_idx = []
    norms_to_calc = {}
    sg_error = False
    ngon_error = False

    # Reminder that obj's use 1 based indexing
    i = 0
    for line in obj_string.split("\n"):
        i += 1
        line = line.lstrip(' ')
        if not line or line[0] == "#":
            continue

        line_parts = line.replace('\t', ' ').split(' ', 1)
        if len(line_parts) < 2:
            continue

        token, line = line_parts

        if token == "v":
            loc_data = [d for d in line.split(' ') if d]
            if len(loc_data) != 3:
                raise ValueError("Invalid vertex coord on line %s." % i)
            v_locs.append(loc_data)

        elif token == "vt":
            uv_data = [d for d in line.split(' ') if d]
            if len(uv_data) not in (2, 3):
                raise ValueError("Invalid vertex texture coord on line %s." % i)
            v_uvs.append(uv_data)

        elif token == "vn":
            norm_data = [d for d in line.split(' ') if d]
            if len(norm_data) != 3:
                raise ValueError("Invalid vertex normal on line %s." % i)
            v_norms.append(norm_data)

        elif token == "f":
            tri_data = [d for d in line.split(' ') if d]
            if len(tri_data) != 3:
                ngon_error = True
                continue

            v_indices = [0, 0, 0]
            for j in range(len(tri_data)):
                v_data = tuple(tri_data[j].replace(' ', '').split('/'))
                if not len(v_data[0]):
                    raise ValueError("Invalid triangle vert description.")

                if v_data not in v_datas:
                    v_data_by_idx.append(v_data)
                    v_datas[v_data] = len(v_data_by_idx)

                v_idx = v_datas[v_data] - 1
                v_indices[j] = v_idx
                if len(v_data) < 3 or not v_data[2]:
                    norms_to_calc.setdefault(v_idx, []).append(f_num)

            tris.append(JmsTriangle(0, mat_num, *v_indices))
            f_num += 1

        elif token == "usemtl":
            name = line
            if name in mat_names:
                mat_num = mat_names[name]
            else:
                mat_num = len(mat_names)
                mats.append(JmsMaterial(name))
                mat_names[name] = mat_num

        elif token == "s" and not sg_error:
            sg_error = True

        else:
            # dont care about anything else
            pass

    if sg_error or ngon_error:
        if ngon_error:
            print("Quads and n-gons are not supported.")

        if sg_error:
            print("Smoothing groups not supported! Export with vertex normals.")

        return

    model.verts = verts = [None] * len(v_data_by_idx)
    for i in range(len(v_data_by_idx)):
        v_data = v_data_by_idx[i]

        loc_idx = int(v_data[0]) - 1
        v_loc = v_locs[loc_idx]
        pos_x = float(v_loc[0])
        pos_y = float(v_loc[1])
        pos_z = float(v_loc[2])

        if len(v_data) > 1 and v_data[1]:
            v_uv = v_uvs[int(v_data[1]) - 1]
            tex_u = float(v_uv[0])
            tex_v = float(v_uv[1])
        else:
            tex_u = pos_x
            tex_v = pos_y

        if len(v_data) > 2 and v_data[2]:
            v_norm = v_norms[int(v_data[2]) - 1]
            norm_i = float(v_norm[0])
            norm_j = float(v_norm[1])
            norm_k = float(v_norm[2])
        else:
            norm_i = 1
            norm_j = norm_k = 0

        # We want to rotate the model, so we don't use the vertex coords
        # exactly as they are given(swap y and z and make z negative).
        verts[i] = JmsVertex(0, pos_x, -pos_z, pos_y,
                             norm_i, -norm_k, norm_j,
                             -1, 0.0, tex_u, tex_v)

    sqrt = math.sqrt
    for i in norms_to_calc:
        vert = verts[i]
        i = j = k = 0
        face_ct = 0

        for f_num in norms_to_calc[i]:
            face_ct += 1
            face = tris[f_num]
            v0 = verts[face.v0]
            v1 = verts[face.v1]
            v2 = verts[face.v2]
            vax = v1.pos_x - v0.pos_x
            vay = v1.pos_y - v0.pos_y
            vaz = v1.pos_z - v0.pos_z

            vbx = v2.pos_x - v0.pos_x
            vby = v2.pos_y - v0.pos_y
            vbz = v2.pos_z - v0.pos_z

            fi = vay * vbz - vaz * vby
            fj = vaz * vbx - vax * vbz
            fk = vax * vby - vay * vbx

            mag = fi**2 + fj**2 + fk**2
            if mag > 0:
                mag = sqrt(mag)
                i += fi / mag
                j += fj / mag
                k += fk / mag

        if face_ct:
            vert.norm_i = i / face_ct
            vert.norm_j = j / face_ct
            vert.norm_k = k / face_ct

    if not mats:
        mats.append(JmsMaterial("__unnamed"))

    return model



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from math import sqrt


def decompress_quaternion48(word_0, word_1, word_2):
    '''Decompress a ones-signed 6byte quaternion to floats'''
    comp_rot = (word_2 & 0xFFff) | ((word_1 & 0xFFff)<<16) | ((word_0 & 0xFFff)<<32)
    w =  comp_rot & 4095
    k = (comp_rot >> 12) & 4095
    j = (comp_rot >> 24) & 4095
    i = (comp_rot >> 36) & 4095
    # avoid division by zero
    if i | j | k | w:
        if i & 0x800: i -= 4095
        if j & 0x800: j -= 4095
        if k & 0x800: k -= 4095
        if w & 0x800: w -= 4095
        length = 1.0 / sqrt(i**2 + j**2 + k**2 + w**2)
        return i * length, j * length, k * length, w * length
    return 0.0, 0.0, 0.0, 1.0


def compress_quaternion48(i, j, k, w):
    '''Compress normalized quaternion to a ones-signed 6byte representation'''
    comp_rot = 0
    nmag = i**2 + j**2 + k**2 + w**2
    # avoid division by zero
    if nmag:
        nmag = 2047 / sqrt(nmag)
        i = int(round(i*nmag)) % 4095
        j = int(round(j*nmag)) % 4095
        k = int(round(k*nmag)) % 4095
        w = int(round(w*nmag)) % 4095
        comp_rot = w | (k << 12) | (j << 24) | (i << 36)
    else:
        comp_rot = 2047

    return comp_rot>>32, (comp_rot>>16) & 0xFFff, comp_rot & 0xFFff


def decompress_normal32(n):
    i = (n&1023) / 1023
    j = ((n>>11)&1023) / 1023
    k = ((n>>22)&511) / 511
    if n&(1<<10): i -= 1.0
    if n&(1<<21): j -= 1.0
    if n&(1<<31): k -= 1.0

    return i, j, k


def compress_normal32(i, j, k):
    i = min(max(i, -1.0), 1.0)
    j = min(max(j, -1.0), 1.0)
    k = min(max(k, -1.0), 1.0)
    # original algorithm before shelly's optimization, kept for clarity
    #if i < 0: i += 2047
    #if j < 0: j += 2047
    #if k < 0: k += 1023
    #return i | (j << 11) | (k << 22)
    return ((int(round(i*1023)) % 2047) |
            ((int(round(j*1023)) % 2047) << 11) |
            ((int(round(k*511)) % 1023) << 22))


#uncomp_norm = [.333, -.75, 1]
#nmag = sqrt(sum(uncomp_norm[i]**2 for i in range(3)))
#uncomp_norm = [val / nmag for val in uncomp_norm]
#comp_norm = compress_normal32(*uncomp_norm)
#print(uncomp_norm)
#print(comp_norm)
#print(decompress_normal32(comp_norm))

#orig = (0xdab8, 0x67f1, 0x2fff)
#recomp = compress_quaternion48(*decompress_quaternion48(*orig))
#print(orig)
#print(recomp)
#print(decompress_quaternion48(*orig))
#print(decompress_quaternion48(*recomp))



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Namings and stuff
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

chgd_hud_global_biped = (
    "spartan",
    "elite",
    "monitor",
    )


chgd_hud_global_hud_attribute = Struct("hud_attribute",
    Bool32("resolution_flags",
        "wide_full",
        "wide_half",
        ("standard_full", 1 << 3),
        "wide_quarter",
        "standard_half",
        ("standard_quarter", 1 << 7),
        ),
    float_rad("warp_angle"),
    Float("warp_amount"),
    Float("warp_direction"),
    UInt32("resolution_width"),
    UInt32("resolution_height"),
    QStruct("motion_sensor_offset", INCLUDE=xy_float),
    Float("motion_sensor_radius"),
    Float("motion_sensor_scale"),
    Float("horizontal_scale"),
    Float("vertical_scale"),
    Float("horizontal_stretch"),
    Float("vertical_stretch"),
    Float("notification_offset_y"),
    Float("state_left_right_offset_y"),
    Float("state_center_offset_y"),
    Float("unknown_0", VISIBLE=False),
    Float("unknown_1", VISIBLE=False),
    Float("unknown_2", VISIBLE=False),
    Float("state_scale"),
    Float("notification_scale"),
    Float("notification_line_spacing"),
    SInt16("unknown_3", VISIBLE=False),
    SInt16("unknown_4", VISIBLE=False),
    ENDIAN=">", SIZE=96
    )


chgd_hud_global_hud_sound = Struct("hud_sound",
    h3_dependency("spartan_sound"),
    Bool32("latched_to",
        "shield_recharging",
        "shield_damaged",
        "shield_low",
        "shield_empty",
        "health_low",
        "health_empty",
        "health_minor_damage",
        "health_major_damage",
        "rocket_locking",
        "rocket_locked",
        "missile_locking",
        "missile_locked",
        ),
    Float("scale"),
    h3_dependency("elite_sound"),
    ENDIAN=">", SIZE=40
    )


chgd_hud_global_multiplayer_medal = Struct("multiplayer_medal",
    h3_string_id("medal"),
    ENDIAN=">", SIZE=4
    )


chgd_hud_global = Struct("hud_global",
    SEnum32("biped", *chgd_hud_global_biped),
    color_argb_uint32("_0_hud_disabled"),
    color_argb_uint32("_1_hud_primary"),
    color_argb_uint32("_2_hud_foreground"),
    color_argb_uint32("_3_hud_warning"),
    color_argb_uint32("_4_neutral_reticule"),
    color_argb_uint32("_5_hostile_reticule"),
    color_argb_uint32("_6_friendly_reticule"),
    color_argb_uint32("_7"),
    color_argb_uint32("_8_neutral_blip"),
    color_argb_uint32("_9_hostile_blip"),
    color_argb_uint32("_10_friendly_player_blip"),
    color_argb_uint32("_11_friendly_ai_blip"),
    color_argb_uint32("_12"),
    color_argb_uint32("_13_waypoint_blip"),
    color_argb_uint32("_14_distant_waypoint_blip"),
    color_argb_uint32("_15_waypoint"),
    color_argb_uint32("_16"),
    color_argb_uint32("_17"),
    color_argb_uint32("_18"),
    color_argb_uint32("_19"),
    color_argb_uint32("_20"),
    color_argb_uint32("_21"),
    color_argb_uint32("_22_text_fade_in"),
    color_argb_uint32("_23"),
    color_argb_uint32("_24"),
    color_argb_uint32("_25"),
    color_argb_uint32("_26"),
    color_argb_uint32("_27"),
    h3_reflexive("hud_attributes", chgd_hud_global_hud_attribute),
    h3_reflexive("hud_sounds", chgd_hud_global_hud_sound),
    h3_dependency("unknown_0"),
    h3_dependency("frag_grenade_swap_sound"),
    h3_dependency("plasma_grenade_swap_sound"),
    h3_dependency("spike_grenade_swap_sound"),
    h3_dependency("firebomb_grenade_swap_sound"),
    h3_dependency("damage_microtexture"),
    h3_dependency("damage_noise"),
    h3_dependency("directional_arrow"),
    h3_dependency("waypoints"),
    h3_dependency("scoreboard_hud"),
    h3_dependency("metagame_scoreboard_hud"),
    h3_dependency("theater_hud"),
    h3_dependency("forge_hud"),
    h3_dependency("hud_strings"),
    h3_dependency("medals"),
    h3_reflexive("multiplayer_medals", chgd_hud_global_multiplayer_medal),
    h3_dependency("medal_hud_animation"),
    h3_dependency("cortana_channel"),
    h3_dependency("unknown_1"),
    h3_dependency("unknown_2"),
    h3_dependency("jammer_response"),
    h3_dependency("jammer_shield_hit"),
    Float("grenade_scematics_spacing"),
    Float("equipment_scematic_offset_y"),
    Float("dual_equipment_scematic_offset_y"),
    Float("unknown_3", VISIBLE=False),
    Float("unknown_4", VISIBLE=False),
    Float("scoreboard_leader_offset_y"),
    BytesRaw("unknown_5", SIZE=4, VISIBLE=False),
    Float("waypoint_scale"),
    ENDIAN=">", SIZE=520
    )


chgd_hud_shader = Struct("hud_shader",
    h3_dependency("vertex_shader"),
    h3_dependency("pixel_shader"),
    ENDIAN=">", SIZE=32
    )


chgd_unknown_0 = Struct("unknown_0",
    BytesRaw("unknown", SIZE=64, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=64
    )


chgd_unknown_1_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=100, VISIBLE=False),
    h3_dependency("sound_0"),
    BytesRaw("unknown_1", SIZE=96, VISIBLE=False),
    h3_dependency("sound_1"),
    VISIBLE=False,
    ENDIAN=">", SIZE=228
    )


chgd_unknown_1 = Struct("unknown_1",
    BytesRaw("unknown_0", SIZE=4, VISIBLE=False),
    h3_reflexive("unknown_1", chgd_unknown_1_unknown_1),
    VISIBLE=False,
    ENDIAN=">", SIZE=16
    )


chgd_player_training_data = Struct("player_training_data",
    h3_string_id("display_string"),
    SInt16("max_display_time"),
    SInt16("display_count"),
    SInt16("disappear_delay"),
    SInt16("redisplay_delay"),
    Float("display_delay"),
    Bool16("flags",
        "not_in_multiplayer",
        ),
    SInt16("unknown", VISIBLE=False),
    ENDIAN=">", SIZE=20
    )


chgd_body = Struct("tagdata",
    h3_reflexive("hud_globals", chgd_hud_global),
    h3_reflexive("hud_shaders", chgd_hud_shader),
    h3_reflexive("unknown_0", chgd_unknown_0),
    h3_reflexive("unknown_1", chgd_unknown_1),
    h3_reflexive("player_training_data", chgd_player_training_data),
    h3_dependency("start_menu_emblems"),
    h3_dependency("campaign_medals"),
    h3_dependency("campaign_medal_hud_animation"),
    Float("campaign_medal_scale"),
    Float("campaign_medal_spacing"),
    QStruct("campaign_medal_offset", INCLUDE=xy_float),
    Float("metagame_scoreboard_top_y"),
    Float("metagame_scoreboard_spacing"),
    h3_dependency("unit_damage_grid"),
    Float("micro_texture_tile_amount"),
    Float("medium_sensor_blip_scale"),
    Float("small_sensor_blip_scale"),
    Float("large_sensor_blip_scale"),
    Float("sensor_blip_glow_amount"),
    Float("sensor_blip_glow_radius"),
    Float("sensor_blip_glow_opacity"),
    h3_dependency("motion_sensor_blip"),
    h3_dependency("birthday_party_effect"),
    h3_dependency("campaign_flood_mask"),
    h3_dependency("campaign_flood_mask_tile"),
    ENDIAN=">", SIZE=240
    )


def get():
    return chgd_def

chgd_def = TagDef("chgd",
    h3_blam_header('chgd'),
    chgd_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["chgd"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.obje import *

obje_attrs = dict(obje_attrs)
obje_attrs[1] = Bool16('flags',
    'does_not_cast_shadow',
    'transparent_self_occlusion',
    'brighter_than_it_should_be',
    'not_a_pathfinding_obstacle',
    'cast_shadow_by_default',
    {NAME: 'xbox_unknown_bit_8', VALUE: 1<<8, VISIBLE: False},
    {NAME: 'xbox_unknown_bit_11', VALUE: 1<<11, VISIBLE: False},
    )

obje_body = Struct('tagdata',
    obje_attrs
    )

def get():
    return obje_def

obje_def = TagDef("obje",
    blam_header('obje'),
    obje_body,

    ext=".object", endian=">", tag_cls=ObjeTag
    )
